package main

import (
	"context"
	"encoding/json"
	"flag"
	"fmt"
	"log"
	"net/url"
	"os"
	"strconv"
	"strings"
	"time"

	"git.autistici.org/ale/autoradio"
	"git.autistici.org/ale/autoradio/client"
	pb "git.autistici.org/ale/autoradio/proto"
	"github.com/gonuts/commander"
	gonutsflag "github.com/gonuts/flag"
	"go.etcd.io/etcd/clientv3"
)

// Format for output of structured data.
var (
	etcdEndpoints = flag.String("etcd", "http://localhost:2379", "Etcd endpoints (comma-separated list of URLs)")
	outputFormat  = flag.String("format", "txt", "Output format for structured data (json, txt)")
)

type optionalValue struct {
	isSet bool
}

func (v *optionalValue) IsSet() bool {
	return v.isSet
}

type stringOptionalValue struct {
	*optionalValue
	value string
}

func newStringOptionalValue() *stringOptionalValue {
	return &stringOptionalValue{optionalValue: &optionalValue{}}
}

func (v *stringOptionalValue) Set(val string) error {
	v.isSet = true
	v.value = val
	return nil
}

func (v *stringOptionalValue) Value() string {
	return v.value
}

func (v *stringOptionalValue) String() string {
	return v.value
}

func (v *stringOptionalValue) Get() interface{} {
	return v.value
}

type intOptionalValue struct {
	*optionalValue
	value int
}

func newIntOptionalValue() *intOptionalValue {
	return &intOptionalValue{optionalValue: &optionalValue{}}
}

func (v *intOptionalValue) Set(val string) error {
	ival, err := strconv.Atoi(val)
	v.value = ival
	v.isSet = true
	return err
}

func (v *intOptionalValue) Value() int {
	return v.value
}

func (v *intOptionalValue) String() string {
	return fmt.Sprintf("%v", v.value)
}

func (v *intOptionalValue) Get() interface{} {
	return v.value
}

type floatOptionalValue struct {
	*optionalValue
	value float64
}

func newFloatOptionalValue() *floatOptionalValue {
	return &floatOptionalValue{optionalValue: &optionalValue{}}
}

func (v *floatOptionalValue) Set(val string) error {
	fval, err := strconv.ParseFloat(val, 64)
	v.value = fval
	v.isSet = true
	return err
}

func (v *floatOptionalValue) Value() float64 {
	return v.value
}

func (v *floatOptionalValue) String() string {
	return fmt.Sprintf("%v", v.value)
}

func (v *floatOptionalValue) Get() interface{} {
	return v.value
}

type hasAddFlags interface {
	AddFlags(*gonutsflag.FlagSet)
}

type commandInterface interface {
	Run([]string)
	Command() *commander.Command
}

type baseCommand struct {
	UsageLine string
	Short     string
	Long      string
}

func (b *baseCommand) Command() *commander.Command {
	return &commander.Command{
		UsageLine: b.UsageLine,
		Short:     b.Short,
		Long:      b.Long,
	}
}

func (b *baseCommand) Run(args []string) {
}

var auClient *client.Client

func getClient() *client.Client {
	if auClient == nil {
		cli, err := clientv3.New(clientv3.Config{
			Endpoints:   strings.Split(*etcdEndpoints, ","),
			DialTimeout: 5 * time.Second,
		})
		if err != nil {
			log.Fatalf("failed to connect to etcd: %v", err)
		}
		auClient = client.New(cli)
	}
	return auClient
}

func setRelay(m *pb.Mount, relayURL string) {
	if relayURL == "" {
		// Randomly generate source credentials.
		m.SourceUsername = autoradio.GenerateUsername(m.Path)
		m.SourcePassword = autoradio.GeneratePassword()
	} else {
		// Validate the given relay URL.
		u, err := url.Parse(relayURL)
		if err != nil {
			log.Fatal(err)
		}
		m.RelayUrl = u.String()
		m.SourceUsername = ""
		m.SourcePassword = ""
	}
}

func printMount(m *pb.Mount) {
	switch *outputFormat {
	case "json":
		s, _ := json.MarshalIndent(m, "", "    ")
		os.Stdout.Write(s)

	case "txt":
		fmt.Printf("path=%s\n", m.Path)
		if m.SourceUsername != "" {
			fmt.Printf("username=%s\npassword=%s\n", m.SourceUsername, m.SourcePassword)
		}
		if m.RelayUrl != "" {
			fmt.Printf("relay_url=%s\n", m.RelayUrl)
		}
		if m.FallbackPath != "" {
			fmt.Printf("fallback_url=%s\n", m.FallbackPath)
		}
		if t := m.TranscodeParams; t != nil {
			fmt.Printf("transcode_source_url=%s\n", t.SourcePath)
			fmt.Printf("transcode_format=%s\n", t.Format)
			fmt.Printf("transcode_bitrate=%d\n", t.BitRate)
			fmt.Printf("transcode_quality=%f\n", t.Quality)
			fmt.Printf("transcode_samplerate=%d\n", t.SampleRate)
		}

	default:
		log.Printf("unsupported output format \"%s\"", *outputFormat)
	}
}

func mountExists(name string, c *client.Client) bool {
	m, _ := c.GetMount(context.Background(), name)
	return m != nil
}

// Create a new mountpoint.
type createMountCommand struct {
	baseCommand
	relay    string
	fallback string
}

func newCreateMountCommand() *createMountCommand {
	return &createMountCommand{
		baseCommand: baseCommand{
			UsageLine: "create-mount <path>",
			Short:     "Create a new mountpoint",
			Long: `
Create a new mountpoint at the specified PATH (which will become the
absolute URL of the stream).
`,
		},
	}
}

func (cmd *createMountCommand) AddFlags(f *gonutsflag.FlagSet) {
	f.StringVar(&cmd.relay, "relay", "", "Upstream URL to relay")
	f.StringVar(&cmd.fallback, "fallback", "", "Fallback stream URL")
}

func (cmd *createMountCommand) Run(args []string) {
	if len(args) != 1 {
		log.Fatal("Wrong number of arguments")
	}
	fmt.Printf("args: %v\n", args)

	path := args[0]
	if !strings.HasPrefix(path, "/") {
		log.Printf("Warning: mountpoint %s does not start with a slash, using /%s instead", path, path)
		path = "/" + path
	}

	// Check if the mount already exists.
	c := getClient()
	if mountExists(path, c) {
		log.Fatal("ERROR: A mount with that name already exists!")
	}

	// Create the new mount and set the relevant fields (depending
	// on the options passed to the command).
	m := &pb.Mount{Path: path}
	setRelay(m, cmd.relay)
	m.FallbackPath = cmd.fallback

	if err := m.Valid(); err != nil {
		log.Fatalf("ERROR: mount configuration is invalid: %v", err)
	}

	if err := c.SetMount(context.Background(), m); err != nil {
		log.Fatalf("ERROR: creating mount: %v", err)
	}

	printMount(m)
}

// Create a submount (transcoded stream).
type createTranscodingMountCommand struct {
	baseCommand

	sourcePath string
	format     string
	quality    float64
	bitRate    int
	sampleRate int
	channels   int
	stereoMode string
	fallback   string
}

func newCreateTranscodingMountCommand() *createTranscodingMountCommand {
	return &createTranscodingMountCommand{
		baseCommand: baseCommand{
			UsageLine: "create-transcoding-mount <path>",
			Short:     "Create a transcoding mount",
			Long: `
Create a new stream that will transcode the parent stream with
different encoding parameters.
`,
		},
	}
}

func (cmd *createTranscodingMountCommand) AddFlags(f *gonutsflag.FlagSet) {
	f.StringVar(&cmd.sourcePath, "source", "", "Source mountpoint")
	f.StringVar(&cmd.format, "codec", "mp3", "Encoding format")
	f.Float64Var(&cmd.quality, "quality", 0, "Quality (for VBR encoders)")
	f.IntVar(&cmd.bitRate, "bitrate", 32, "Bitrate (Kbps)")
	f.IntVar(&cmd.sampleRate, "samplerate", 44100, "Sample rate (Hz)")
	f.IntVar(&cmd.channels, "channels", 2, "Number of channels")
	f.StringVar(&cmd.stereoMode, "stereo-mode", "joint_stereo", "Stereo mode for mp3 codec (stereo, joint_stereo)")
	f.StringVar(&cmd.fallback, "fallback", "", "Fallback stream URL")
}

func (cmd *createTranscodingMountCommand) Run(args []string) {
	if len(args) != 1 {
		log.Fatal("Wrong number of arguments")
	}

	path := args[0]
	if !strings.HasPrefix(path, "/") {
		log.Printf("Warning: mountpoint %s does not start with a slash, using /%s instead", path, path)
		path = "/" + path
	}

	// The mount path should not exist.
	c := getClient()
	if mountExists(path, c) {
		log.Fatal("ERROR: a mount with that name already exists!")
	}
	// The source mount should exist.
	if !mountExists(cmd.sourcePath, c) {
		log.Fatal("ERROR: the source mount does not exist!")
	}

	// Retrieve the parent mount point and add a TranscodingMount.
	m := &pb.Mount{
		Path:         path,
		FallbackPath: cmd.fallback,
		Transcode:    true,
		TranscodeParams: &pb.EncodingParams{
			SourcePath: cmd.sourcePath,
			Format:     cmd.format,
			BitRate:    int32(cmd.bitRate),
			SampleRate: int32(cmd.sampleRate),
			Channels:   int32(cmd.channels),
			StereoMode: cmd.stereoMode,
			Quality:    float32(cmd.quality),
		},
	}
	setRelay(m, "")

	if err := m.Valid(); err != nil {
		log.Fatalf("ERROR: mount configuration is invalid: %v", err)
	}

	if err := c.SetMount(context.Background(), m); err != nil {
		log.Fatalf("ERROR: creating mount: %v", err)
	}

	printMount(m)
}

// Edit a mountpoint.
type editMountCommand struct {
	baseCommand
	relay           *stringOptionalValue
	fallback        *stringOptionalValue
	transFormat     *stringOptionalValue
	transSource     *stringOptionalValue
	transBitRate    *intOptionalValue
	transSampleRate *intOptionalValue
	transChannels   *intOptionalValue
	transStereoMode *stringOptionalValue
	transQuality    *floatOptionalValue
}

var UNSET = "UNSET"

func newEditMountCommand() *editMountCommand {
	return &editMountCommand{
		baseCommand: baseCommand{
			UsageLine: "edit-mount <path>",
			Short:     "Edit an existing mountpoint",
			Long: `
Modify parameters of the specified mountpoint, such as the relay
and the fallback URL. If the relay option is set, the mountpoint
will not accept source connections anymore. To revert to the
default, non-relay behavior, set the relay to the empty string
(with --relay="").
`,
		},
		relay:           newStringOptionalValue(),
		fallback:        newStringOptionalValue(),
		transFormat:     newStringOptionalValue(),
		transSource:     newStringOptionalValue(),
		transBitRate:    newIntOptionalValue(),
		transSampleRate: newIntOptionalValue(),
		transChannels:   newIntOptionalValue(),
		transStereoMode: newStringOptionalValue(),
		transQuality:    newFloatOptionalValue(),
	}
}

func (cmd *editMountCommand) AddFlags(f *gonutsflag.FlagSet) {
	f.Var(cmd.relay, "relay", "Upstream URL to relay")
	f.Var(cmd.fallback, "fallback", "Fallback stream URL")

	f.Var(cmd.transSource, "source", "[transcoding] Source mountpoint")
	f.Var(cmd.transFormat, "codec", "[transcoding] Encoding format")
	f.Var(cmd.transQuality, "quality", "[transcoding] Quality (for VBR encoders)")
	f.Var(cmd.transBitRate, "bitrate", "[transcoding] Bitrate (Kbps)")
	f.Var(cmd.transSampleRate, "samplerate", "[transcoding] Sample rate (Hz)")
	f.Var(cmd.transChannels, "channels", "[transcoding] Number of channels")
	f.Var(cmd.transStereoMode, "stereo-mode", "[transcoding] Stereo mode for mp3 encoding (stereo, joint_stereo)")
}

func (cmd *editMountCommand) transcodingOptionsSet() bool {
	opts := []interface {
		IsSet() bool
	}{
		cmd.transSource, cmd.transFormat, cmd.transQuality,
		cmd.transBitRate, cmd.transSampleRate, cmd.transChannels,
		cmd.transStereoMode,
	}
	for _, o := range opts {
		if o.IsSet() {
			return true
		}
	}
	return false
}

func (cmd *editMountCommand) Run(args []string) {
	if len(args) != 1 {
		log.Fatal("Wrong number of arguments")
	}

	c := getClient()
	m, err := c.GetMount(context.Background(), args[0])
	if err != nil {
		log.Fatalf("ERROR: %v", err)
	}
	if m == nil {
		log.Fatal("ERROR: mount not found")
	}

	// Only set those fields that were passed on the command line.
	if cmd.fallback.IsSet() {
		m.FallbackPath = cmd.fallback.Value()
	}
	if cmd.relay.IsSet() {
		setRelay(m, cmd.relay.Value())
	}

	if cmd.transcodingOptionsSet() && !m.Transcode {
		log.Fatal("ERROR: can't set transcoding options on a non-transcoding mount (delete and re-create)")
	}

	if cmd.transFormat.IsSet() {
		m.TranscodeParams.Format = cmd.transFormat.Value()
	}
	if cmd.transSource.IsSet() {
		m.TranscodeParams.SourcePath = cmd.transSource.Value()
	}
	if cmd.transBitRate.IsSet() {
		m.TranscodeParams.BitRate = int32(cmd.transBitRate.Value())
	}
	if cmd.transSampleRate.IsSet() {
		m.TranscodeParams.SampleRate = int32(cmd.transSampleRate.Value())
	}
	if cmd.transQuality.IsSet() {
		m.TranscodeParams.Quality = float32(cmd.transQuality.Value())
	}
	if cmd.transChannels.IsSet() {
		m.TranscodeParams.Channels = int32(cmd.transChannels.Value())
	}
	if cmd.transStereoMode.IsSet() {
		m.TranscodeParams.StereoMode = cmd.transStereoMode.Value()
	}

	if err := m.Valid(); err != nil {
		log.Fatalf("ERROR: mount configuration is invalid: %v", err)
	}

	if err := c.SetMount(context.Background(), m); err != nil {
		log.Fatalf("ERROR: updating mount: %v", err)
	}

	printMount(m)
}

// Delete an existing mountpoint.
type deleteMountCommand struct {
	baseCommand
}

func newDeleteMountCommand() *deleteMountCommand {
	return &deleteMountCommand{
		baseCommand{
			UsageLine: "delete-mount <path>",
			Short:     "Delete a mountpoint",
			Long: `
Delete the specified mountpoint.
`,
		},
	}
}

func (cmd *deleteMountCommand) Run(args []string) {
	if len(args) != 1 {
		log.Fatal("Wrong number of arguments")
	}
	path := args[0]
	c := getClient()
	if !mountExists(path, c) {
		log.Fatal("ERROR: mount not found")
	}

	if err := c.DeleteMount(context.Background(), path); err != nil {
		log.Fatalf("ERROR: deleting mount: %v", err)
	}

	// Delete all the transcoding mounts that have this as a
	// source.
	mounts, err := c.ListMounts(context.Background())
	if err != nil {
		log.Fatalf("ERROR: %v", err)
	}
	for _, m := range mounts {
		if m.HasTranscoder() && m.TranscodeParams.SourcePath == path {
			if err := c.DeleteMount(context.Background(), m.Path); err != nil {
				log.Printf("ERROR: deleting transcoded mount %s: %v", m.Path, err)
			}
		}
	}

	log.Printf("mountpoint %s removed", path)
}

// List known mountpoints.
type listMountsCommand struct {
	baseCommand
}

func newListMountsCommand() *listMountsCommand {
	return &listMountsCommand{
		baseCommand{
			UsageLine: "list-mounts",
			Short:     "List all configured mountpoints",
			Long: `
Outputs a list of all the currently configured mountpoints.
`,
		},
	}
}

func (cmd *listMountsCommand) Run(args []string) {
	if len(args) != 0 {
		log.Fatal("Too many arguments")
	}

	mounts, err := getClient().ListMounts(context.Background())
	if err != nil {
		log.Fatalf("ERROR: %v", err)
	}
	var names []string
	for _, m := range mounts {
		names = append(names, m.Path)
	}

	switch *outputFormat {
	case "json":
		s, _ := json.MarshalIndent(names, "", "    ")
		os.Stdout.Write(s)

	//case "txt":
	default:
		for _, n := range names {
			fmt.Println(n)
		}
	}
}

// Show mountpoint information.
type showMountCommand struct {
	baseCommand
}

func newShowMountCommand() *showMountCommand {
	return &showMountCommand{
		baseCommand{
			UsageLine: "show-mount <path>",
			Short:     "Show mountpoint information",
			Long: `
Print information about the specified mountpoint (including
the source credentials).
`,
		},
	}
}

func (cmd *showMountCommand) Run(args []string) {
	if len(args) != 1 {
		log.Fatal("Wrong number of arguments")
	}

	m, err := getClient().GetMount(context.Background(), args[0])
	if err != nil {
		log.Fatalf("ERROR: %v", err)
	}
	if m == nil {
		log.Fatal("ERROR: mount not found")
	}

	printMount(m)
}

// Backup mount configuration.
type backupCommand struct {
	baseCommand
}

func newBackupCommand() *backupCommand {
	return &backupCommand{
		baseCommand{
			UsageLine: "backup",
			Short:     "Backup mount configuration",
			Long: `
Dump the autoradio configuration to stdout, in a format that is
understood by the "restore" command.
`,
		},
	}
}

func (cmd *backupCommand) Run(args []string) {
	if len(args) != 0 {
		log.Fatal("Too many arguments")
	}

	mounts, err := getClient().ListMounts(context.Background())
	if err != nil {
		log.Fatalf("ERROR: %v", err)
	}
	if err := json.NewEncoder(os.Stdout).Encode(mounts); err != nil {
		log.Fatalf("ERROR: %v", err)
	}
}

// Restore mount configuration.
type restoreCommand struct {
	baseCommand
}

func newRestoreCommand() *restoreCommand {
	return &restoreCommand{
		baseCommand{
			UsageLine: "restore",
			Short:     "Restore mount configuration",
			Long: `
Read a configuration dump from standard input and restore it.
`,
		},
	}
}

func (cmd *restoreCommand) Run(args []string) {
	if len(args) != 0 {
		log.Fatal("Too many arguments")
	}

	var mounts []*pb.Mount
	if err := json.NewDecoder(os.Stdin).Decode(&mounts); err != nil {
		log.Fatalf("ERROR: %v", err)
	}

	c := getClient()
	for _, m := range mounts {
		if err := c.SetMount(context.Background(), m); err != nil {
			log.Printf("ERROR: creating mount %s: %v", m.Path, err)
		}
	}
}

var cmdr = &commander.Command{
	UsageLine: "radioctl <command> [<args>...]",
	Short:     "Manage `autoradio' configuration",
}

func addCommand(c commandInterface) {
	cmd := c.Command()
	cmd.Run = func(unused *commander.Command, args []string) error {
		c.Run(args)
		return nil
	}
	if afc, ok := c.(hasAddFlags); ok {
		afc.AddFlags(&cmd.Flag)
	}
	cmdr.Subcommands = append(cmdr.Subcommands, cmd)
}

func init() {
	addCommand(newCreateMountCommand())
	addCommand(newCreateTranscodingMountCommand())
	addCommand(newEditMountCommand())
	addCommand(newDeleteMountCommand())
	addCommand(newListMountsCommand())
	addCommand(newShowMountCommand())
	addCommand(newBackupCommand())
	addCommand(newRestoreCommand())
	flag.Usage = usage
}

func usage() {
	fmt.Fprintf(os.Stderr, "Usage: %s\n\nGlobal options:\n", cmdr.UsageLine)
	flag.PrintDefaults()
	fmt.Fprintf(os.Stderr, "\nType \"radioctl help\" for more documentation.\n\n")
}

func main() {
	flag.Parse()
	log.SetFlags(0)
	if err := cmdr.Dispatch(flag.Args()); err != nil {
		log.Fatal(err)
	}
}
